-- viz type=table, name=Table estimates, columns=table_name,object_type,table_format,estimated_hours
-- widget title=Table estimates, row=3, col=2, size_x=3, size_y=8
SELECT * FROM $inventory.table_estimates
