-- viz type=table, name=UDF Summary, search_by=name, columns=catalog,database,name,func_type,func_input,func_returns,deterministic,data_access,body,comment
-- widget title=UDF Summary, row=14, col=0, size_x=8, size_y=8
SELECT * FROM $inventory.udfs
